import hashlib
import os
import re
from http.server import BaseHTTPRequestHandler
from tempfile import TemporaryDirectory

from kagglehub.clients import KaggleApiV1Client
from kagglehub.exceptions import DataCorruptionError
from kagglehub.integrity import to_b64_digest
from tests.fixtures import BaseTestCase

from .utils import create_test_http_server, get_test_file_path


class KaggleAPIHandler(BaseHTTPRequestHandler):
    def do_HEAD(self) -> None:  # noqa: N802
        self.send_response(200)

    def do_GET(self) -> None:  # noqa: N802
        test_file_path = get_test_file_path("foo.txt")
        with open(test_file_path, "rb") as f:
            if self.path.endswith("no-integrity"):
                # Do not set the "x-goog-hash" header which skips integrity check.
                self.send_response(200)
                self.send_header("Content-type", "application/octet-stream")
                self.send_header("Content-Length", str(os.path.getsize(test_file_path)))
                self.end_headers()
                self.wfile.write(f.read())
            if self.path.endswith("corrupted"):
                self.send_response(200)
                self.send_header("Content-type", "application/octet-stream")
                self.send_header("Content-Length", str(os.path.getsize(test_file_path)))
                self.send_header("x-goog-hash", "md5=badhash")
                self.end_headers()
                self.wfile.write(f.read())
            if self.path.endswith("good"):
                start = 0
                if "Range" in self.headers:
                    bytes_range = self.headers["Range"]
                    # See: https://developer.mozilla.org/en-US/docs/Web/HTTP/Range_requests
                    # Format is bytes=<start>-<end>
                    # If <end> is missing, then, it means until the end.
                    # This test server only supports bytes=<start>- format since it is the only one we need.
                    m = re.match("^bytes=([0-9]+)-$", bytes_range)
                    if not m:
                        self.end_headers()
                        self.send_response(400)  # Bad Request
                        return
                    start = int(m.group(1))

                f.seek(start)
                self.send_response(200)
                self.send_header("Content-type", "application/octet-stream")
                self.send_header("Content-Length", str(os.path.getsize(test_file_path) - start))
                self.send_header("Accept-Ranges", "bytes")  # support resumable download

                content = f.read()
                file_hash = hashlib.md5()
                file_hash.update(content)
                # The crc32c hash is set to a random value and is ignored by kagglehub which only checks the md5.
                self.send_header("x-goog-hash", f"crc32c=n03x6A==, md5={to_b64_digest(file_hash)}")
                self.end_headers()
                self.wfile.write(content)

            else:
                self.send_response(404)
                self.wfile.write(bytes(f"Unhandled path: {self.path}", "utf-8"))


class TestKaggleApiV1Client(BaseTestCase):
    def test_download_with_integrity_check(self) -> None:
        with create_test_http_server(KaggleAPIHandler):
            with TemporaryDirectory() as d:
                out_file = os.path.join(d, "out")

                api_client = KaggleApiV1Client()
                api_client.download_file("good", out_file)

                with open(out_file) as f:
                    self.assertEqual("foo", f.read())

    def test_resumable_download_with_integrity_check(self) -> None:
        with create_test_http_server(KaggleAPIHandler):
            with TemporaryDirectory() as d:
                out_file = os.path.join(d, "out")

                # If the out_file already has data, we use the 'Range' header to resume download.
                with open(out_file, "w") as f:
                    f.write("fo")  # Should download the remaining "o".

                api_client = KaggleApiV1Client()
                with self.assertLogs("kagglehub", level="INFO") as cm:
                    api_client.download_file("good", out_file)
                    self.assertIn("INFO:kagglehub.clients:Resuming download from 2 bytes (1 bytes left)...", cm.output)

                with open(out_file) as f:
                    self.assertEqual("foo", f.read())

    def test_download_no_integrity_check(self) -> None:
        with create_test_http_server(KaggleAPIHandler):
            with TemporaryDirectory() as d:
                out_file = os.path.join(d, "out")

                api_client = KaggleApiV1Client()
                api_client.download_file("no-integrity", out_file)

                with open(out_file) as f:
                    self.assertEqual("foo", f.read())

    def test_download_corrupted_file_fail_integrity_check(self) -> None:
        with create_test_http_server(KaggleAPIHandler):
            with TemporaryDirectory() as d:
                out_file = os.path.join(d, "out")

                api_client = KaggleApiV1Client()
                with self.assertRaises(DataCorruptionError):
                    api_client.download_file("corrupted", out_file)

                # Assert the corrupted file has been deleted.
                self.assertFalse(os.path.exists(out_file))



import hashlib
import json
import os
from http.server import BaseHTTPRequestHandler

import requests

import kagglehub
from kagglehub.cache import MODELS_CACHE_SUBFOLDER, get_cached_archive_path
from kagglehub.handle import parse_model_handle
from kagglehub.http_resolver import MODEL_INSTANCE_VERSION_FIELD
from kagglehub.integrity import GCS_HASH_HEADER, to_b64_digest
from tests.fixtures import BaseTestCase

from .utils import create_test_cache, create_test_http_server, get_test_file_path

INVALID_ARCHIVE_MODEL_HANDLE = "metaresearch/llama-2/pyTorch/bad-archive-variation/1"
VERSIONED_MODEL_HANDLE = "metaresearch/llama-2/pyTorch/13b/3"
UNVERSIONED_MODEL_HANDLE = "metaresearch/llama-2/pyTorch/13b"
TEST_FILEPATH = "config.json"
TEST_CONTENTS = "{}"


class KaggleAPIHandler(BaseHTTPRequestHandler):
    def do_HEAD(self) -> None:  # noqa: N802
        self.send_response(200)

    def do_GET(self) -> None:  # noqa: N802
        if self.path.endswith(VERSIONED_MODEL_HANDLE + "/download/" + TEST_FILEPATH):
            test_file_path = get_test_file_path(TEST_FILEPATH)
            with open(test_file_path, "rb") as f:
                # Serve a single file
                self.send_response(200)
                self.send_header("Content-type", "application/octet-stream")
                self.send_header("Content-Length", str(os.path.getsize(test_file_path)))

                content = f.read()
                file_hash = hashlib.md5()
                file_hash.update(content)
                self.send_header(GCS_HASH_HEADER, f"md5={to_b64_digest(file_hash)}")
                self.end_headers()
                self.wfile.write(content)
        elif self.path.endswith(VERSIONED_MODEL_HANDLE + "/download"):
            test_file_path = get_test_file_path("archive.tar.gz")
            with open(test_file_path, "rb") as f:
                # Serve archive file
                self.send_response(200)
                self.send_header("Content-type", "application/x-gzip")
                self.send_header("Content-Length", str(os.path.getsize(test_file_path)))
                content = f.read()
                file_hash = hashlib.md5()
                file_hash.update(content)
                self.send_header(GCS_HASH_HEADER, f"md5={to_b64_digest(file_hash)}")
                self.end_headers()
                self.wfile.write(content)
        elif self.path.endswith(INVALID_ARCHIVE_MODEL_HANDLE + "/download"):
            # Serve a bad archive file
            content = b"bad archive"
            self.send_response(200)
            self.send_header("Content-type", "application/x-gzip")
            self.send_header("Content-Length", str(len(content)))
            self.end_headers()
            self.wfile.write(content)  # bad archive
        elif self.path.endswith(UNVERSIONED_MODEL_HANDLE + "/get"):
            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(bytes(json.dumps({MODEL_INSTANCE_VERSION_FIELD: 3}), "utf-8"))
        else:
            self.send_response(404)
            self.wfile.write(bytes(f"Unhandled path: {self.path}", "utf-8"))


EXPECTED_MODEL_SUBDIR = os.path.join(MODELS_CACHE_SUBFOLDER, "metaresearch", "llama-2", "pyTorch", "13b", "3")
EXPECTED_MODEL_SUBPATH = os.path.join(
    MODELS_CACHE_SUBFOLDER,
    "metaresearch",
    "llama-2",
    "pyTorch",
    "13b",
    "3",
    TEST_FILEPATH,
)


# Test cases for the ModelHttpResolver.
class TestHttpModelDownload(BaseTestCase):
    def _download_model_and_assert_downloaded(
        self, d: str, model_handle: str, expected_subdir_or_subpath: str, **kwargs  # noqa: ANN003
    ) -> None:
        # Download the full model and ensure all files are there.
        model_path = kagglehub.model_download(model_handle, **kwargs)
        self.assertEqual(os.path.join(d, expected_subdir_or_subpath), model_path)
        self.assertEqual(["config.json", "model.keras"], sorted(os.listdir(model_path)))

        # Assert that the archive file has been deleted.
        archive_path = get_cached_archive_path(parse_model_handle(model_handle))
        self.assertFalse(os.path.exists(archive_path))

    def _download_test_file_and_assert_downloaded(self, d: str, model_handle: str, **kwargs) -> None:  # noqa: ANN003
        model_path = kagglehub.model_download(model_handle, path=TEST_FILEPATH, **kwargs)
        self.assertEqual(os.path.join(d, EXPECTED_MODEL_SUBPATH), model_path)
        with open(model_path) as model_file:
            self.assertEqual(TEST_CONTENTS, model_file.readline())

    def test_unversioned_model_download(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                self._download_model_and_assert_downloaded(d, UNVERSIONED_MODEL_HANDLE, EXPECTED_MODEL_SUBDIR)

    def test_versioned_model_download(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                self._download_model_and_assert_downloaded(d, VERSIONED_MODEL_HANDLE, EXPECTED_MODEL_SUBDIR)

    def test_versioned_model_full_download_with_file_already_cached(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                # Download a single file first
                kagglehub.model_download(VERSIONED_MODEL_HANDLE, path=TEST_FILEPATH)
                self._download_model_and_assert_downloaded(d, VERSIONED_MODEL_HANDLE, EXPECTED_MODEL_SUBDIR)

    def test_unversioned_model_full_download_with_file_already_cached(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                # Download a single file first
                kagglehub.model_download(UNVERSIONED_MODEL_HANDLE, path=TEST_FILEPATH)
                self._download_model_and_assert_downloaded(d, UNVERSIONED_MODEL_HANDLE, EXPECTED_MODEL_SUBDIR)

    def test_unversioned_model_download_with_force_download(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                self._download_model_and_assert_downloaded(
                    d, UNVERSIONED_MODEL_HANDLE, EXPECTED_MODEL_SUBDIR, force_download=True
                )

    def test_versioned_model_download_with_force_download(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                self._download_model_and_assert_downloaded(
                    d, VERSIONED_MODEL_HANDLE, EXPECTED_MODEL_SUBDIR, force_download=True
                )

    def test_versioned_model_full_download_with_file_already_cached_and_force_download(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                # Download a single file first
                kagglehub.model_download(VERSIONED_MODEL_HANDLE, path=TEST_FILEPATH)
                self._download_model_and_assert_downloaded(
                    d, VERSIONED_MODEL_HANDLE, EXPECTED_MODEL_SUBDIR, force_download=True
                )

    def test_unversioned_model_full_download_with_file_already_cached_and_force_download(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                # Download a single file first
                kagglehub.model_download(UNVERSIONED_MODEL_HANDLE, path=TEST_FILEPATH)
                self._download_model_and_assert_downloaded(
                    d, UNVERSIONED_MODEL_HANDLE, EXPECTED_MODEL_SUBDIR, force_download=True
                )

    def test_versioned_model_download_bad_archive(self) -> None:
        with create_test_cache():
            with create_test_http_server(KaggleAPIHandler):
                with self.assertRaises(ValueError):
                    kagglehub.model_download(INVALID_ARCHIVE_MODEL_HANDLE)

    def test_versioned_model_download_with_path(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                self._download_test_file_and_assert_downloaded(d, VERSIONED_MODEL_HANDLE)

    def test_unversioned_model_download_with_path(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                self._download_test_file_and_assert_downloaded(d, UNVERSIONED_MODEL_HANDLE)

    def test_versioned_model_download_with_path_with_force_download(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                self._download_test_file_and_assert_downloaded(d, VERSIONED_MODEL_HANDLE, force_download=True)

    def test_unversioned_model_download_with_path_with_force_download(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                self._download_test_file_and_assert_downloaded(d, UNVERSIONED_MODEL_HANDLE, force_download=True)

    def test_versioned_model_download_already_cached(self) -> None:
        with create_test_cache() as d:
            # Download from server.
            with create_test_http_server(KaggleAPIHandler):
                kagglehub.model_download(VERSIONED_MODEL_HANDLE)

            # No internet, cache hit.
            model_path = kagglehub.model_download(VERSIONED_MODEL_HANDLE)

            self.assertEqual(os.path.join(d, EXPECTED_MODEL_SUBDIR), model_path)

    def test_versioned_model_download_with_path_already_cached(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                kagglehub.model_download(VERSIONED_MODEL_HANDLE, path=TEST_FILEPATH)

            # No internet, cache hit.
            model_path = kagglehub.model_download(VERSIONED_MODEL_HANDLE, path=TEST_FILEPATH)

            self.assertEqual(os.path.join(d, EXPECTED_MODEL_SUBPATH), model_path)

    def test_versioned_model_download_already_cached_with_force_download(self) -> None:
        with create_test_cache():
            with create_test_http_server(KaggleAPIHandler):
                kagglehub.model_download(VERSIONED_MODEL_HANDLE)

            # No internet should throw an error.
            with self.assertRaises(requests.exceptions.ConnectionError):
                kagglehub.model_download(VERSIONED_MODEL_HANDLE, force_download=True)

    def test_versioned_model_download_with_path_already_cached_with_force_download(self) -> None:
        with create_test_cache():
            with create_test_http_server(KaggleAPIHandler):
                kagglehub.model_download(VERSIONED_MODEL_HANDLE, path=TEST_FILEPATH)

            # No internet should throw an error.
            with self.assertRaises(requests.exceptions.ConnectionError):
                kagglehub.model_download(VERSIONED_MODEL_HANDLE, path=TEST_FILEPATH, force_download=True)

    def test_versioned_model_download_already_cached_with_force_download_explicit_false(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                kagglehub.model_download(VERSIONED_MODEL_HANDLE)

            # Not force downloaded, cache hit.
            model_path = kagglehub.model_download(VERSIONED_MODEL_HANDLE, force_download=False)

            self.assertEqual(os.path.join(d, EXPECTED_MODEL_SUBDIR), model_path)

    def test_versioned_model_download_with_path_already_cached_with_force_download_explicit_false(self) -> None:
        with create_test_cache() as d:
            with create_test_http_server(KaggleAPIHandler):
                kagglehub.model_download(VERSIONED_MODEL_HANDLE, path=TEST_FILEPATH)

            # Not force downloaded, cache hit.
            model_path = kagglehub.model_download(VERSIONED_MODEL_HANDLE, path=TEST_FILEPATH, force_download=False)

            self.assertEqual(os.path.join(d, EXPECTED_MODEL_SUBPATH), model_path)



import os
from pathlib import Path

from kagglehub.cache import (
    MODELS_CACHE_SUBFOLDER,
    delete_from_cache,
    get_cached_archive_path,
    get_cached_path,
    load_from_cache,
    mark_as_complete,
    mark_as_incomplete,
)
from kagglehub.handle import ModelHandle
from tests.fixtures import BaseTestCase

from .utils import InvalidResourceHandle, create_test_cache

EXPECTED_MODEL_SUBDIR = os.path.join(
    MODELS_CACHE_SUBFOLDER, "google", "bert", "tensorFlow2", "answer-equivalence-bem", "2"
)

EXPECTED_MODEL_SUBPATH = os.path.join(
    MODELS_CACHE_SUBFOLDER,
    "google",
    "bert",
    "tensorFlow2",
    "answer-equivalence-bem",
    "2",
    "foo.txt",
)

TEST_MODEL_HANDLE = ModelHandle(
    owner="google",
    model="bert",
    framework="tensorFlow2",
    variation="answer-equivalence-bem",
    version=2,
)

TEST_FILEPATH = "foo.txt"
TEST_MODEL_VARIABLES_DIR_NAME = "variables"
TEST_MODEL_VARIABLES_FILE_NAME = "variables.txt"


class TestCache(BaseTestCase):
    def test_load_from_cache_miss(self) -> None:
        self.assertEqual(None, load_from_cache(TEST_MODEL_HANDLE))

    def test_load_from_cache_with_path_miss(self) -> None:
        self.assertEqual(None, load_from_cache(TEST_MODEL_HANDLE, TEST_FILEPATH))

    def test_load_from_cache_not_complete_miss(self) -> None:
        with create_test_cache():
            cache_path = get_cached_path(TEST_MODEL_HANDLE)

            # Should be a cache `miss` if the directory exists but not the marker file.
            os.makedirs(cache_path)

            self.assertEqual(None, load_from_cache(TEST_MODEL_HANDLE))

    def test_load_from_cache_with_path_not_complete_miss(self) -> None:
        with create_test_cache():
            cache_path = get_cached_path(TEST_MODEL_HANDLE)

            # Should be a cache `miss` if the directory and file exist but not the marker file.
            os.makedirs(cache_path)
            Path(os.path.join(cache_path, TEST_FILEPATH)).touch()  # Create file

            self.assertEqual(None, load_from_cache(TEST_MODEL_HANDLE, path=TEST_FILEPATH))

    def test_load_from_cache_with_complete_marker_no_files_miss(self) -> None:
        with create_test_cache():
            # Should be a cache `miss` if completion marker file exists but not the files themselves.
            mark_as_complete(TEST_MODEL_HANDLE)

            self.assertEqual(None, load_from_cache(TEST_MODEL_HANDLE))

    def test_load_from_cache_with_path_complete_marker_no_files_miss(self) -> None:
        with create_test_cache():
            # Should be a cache `miss` if completion marker file exists but not the file itself.
            mark_as_complete(TEST_MODEL_HANDLE, path=TEST_FILEPATH)

            self.assertEqual(None, load_from_cache(TEST_MODEL_HANDLE, path=TEST_FILEPATH))

    def test_cache_hit(self) -> None:
        with create_test_cache() as d:
            cache_path = get_cached_path(TEST_MODEL_HANDLE)
            os.makedirs(cache_path)
            mark_as_complete(TEST_MODEL_HANDLE)

            path = load_from_cache(TEST_MODEL_HANDLE)

            self.assertEqual(
                os.path.join(d, EXPECTED_MODEL_SUBDIR),
                path,
            )

    def test_cache_hit_with_path(self) -> None:
        with create_test_cache() as d:
            cache_path = get_cached_path(TEST_MODEL_HANDLE)

            os.makedirs(cache_path)
            Path(os.path.join(cache_path, TEST_FILEPATH)).touch()  # Create file
            mark_as_complete(TEST_MODEL_HANDLE, path=TEST_FILEPATH)

            path = load_from_cache(TEST_MODEL_HANDLE, path=TEST_FILEPATH)

            self.assertEqual(
                os.path.join(d, EXPECTED_MODEL_SUBPATH),
                path,
            )

    def test_load_from_cache_invalid_handle(self) -> None:
        with self.assertRaises(ValueError):
            load_from_cache(InvalidResourceHandle())

    def test_model_archive_path(self) -> None:
        with create_test_cache() as d:
            archive_path = get_cached_archive_path(TEST_MODEL_HANDLE)

            self.assertEqual(
                os.path.join(
                    d,
                    MODELS_CACHE_SUBFOLDER,
                    "google",
                    "bert",
                    "tensorFlow2",
                    "answer-equivalence-bem",
                    "2.archive",
                ),
                archive_path,
            )

    def _download_test_model_to_cache(self) -> None:
        cache_path = get_cached_path(TEST_MODEL_HANDLE)
        model_variable_dir = os.path.join(cache_path, TEST_MODEL_VARIABLES_DIR_NAME)

        os.makedirs(model_variable_dir)
        Path(os.path.join(cache_path, TEST_FILEPATH)).touch()
        Path(os.path.join(model_variable_dir, TEST_MODEL_VARIABLES_FILE_NAME)).touch()

        mark_as_complete(TEST_MODEL_HANDLE)

    def _download_test_file_to_cache(self) -> None:
        cache_path = get_cached_path(TEST_MODEL_HANDLE)

        os.makedirs(cache_path)
        Path(os.path.join(cache_path, TEST_FILEPATH)).touch()

        mark_as_complete(TEST_MODEL_HANDLE, path=TEST_FILEPATH)

    def test_delete_from_cache(self) -> None:
        with create_test_cache() as d:
            self._download_test_model_to_cache()

            deleted_path = delete_from_cache(TEST_MODEL_HANDLE)

            self.assertEqual(os.path.join(d, EXPECTED_MODEL_SUBDIR), deleted_path)
            self.assertFalse(os.path.exists(get_cached_path(TEST_MODEL_HANDLE)))

    def test_delete_from_cache_with_path(self) -> None:
        with create_test_cache() as d:
            self._download_test_file_to_cache()

            deleted_path = delete_from_cache(TEST_MODEL_HANDLE, path=TEST_FILEPATH)

            self.assertEqual(os.path.join(d, EXPECTED_MODEL_SUBPATH), deleted_path)
            self.assertFalse(os.path.exists(os.path.join(get_cached_path(TEST_MODEL_HANDLE), TEST_FILEPATH)))

    def test_delete_from_cache_without_files_without_complete_marker(self) -> None:
        with create_test_cache():
            deleted_path = delete_from_cache(TEST_MODEL_HANDLE)
            self.assertEqual(None, deleted_path)

    def test_delete_from_cache_without_files_without_complete_marker_with_path(self) -> None:
        with create_test_cache():
            deleted_path = delete_from_cache(TEST_MODEL_HANDLE, path=TEST_FILEPATH)
            self.assertEqual(None, deleted_path)

    def test_delete_from_cache_without_files_with_complete_marker(self) -> None:
        with create_test_cache():
            mark_as_complete(TEST_MODEL_HANDLE)

            deleted_path = delete_from_cache(TEST_MODEL_HANDLE)

            # Should not delete anything if only the marker file existed.
            self.assertEqual(None, deleted_path)

    def test_delete_from_cache_without_files_with_complete_marker_with_path(self) -> None:
        with create_test_cache():
            mark_as_complete(TEST_MODEL_HANDLE, path=TEST_FILEPATH)

            deleted_path = delete_from_cache(TEST_MODEL_HANDLE, path=TEST_FILEPATH)

            # Should not delete anything if only the marker file existed.
            self.assertEqual(None, deleted_path)

    def test_delete_from_cache_with_files_without_complete_marker(self) -> None:
        with create_test_cache() as d:
            self._download_test_model_to_cache()
            mark_as_incomplete(TEST_MODEL_HANDLE)

            deleted_path = delete_from_cache(TEST_MODEL_HANDLE)

            self.assertEqual(os.path.join(d, EXPECTED_MODEL_SUBDIR), deleted_path)
            self.assertFalse(os.path.exists(get_cached_path(TEST_MODEL_HANDLE)))

    def test_delete_from_cache_with_files_without_complete_marker_with_path(self) -> None:
        with create_test_cache() as d:
            self._download_test_file_to_cache()
            mark_as_incomplete(TEST_MODEL_HANDLE, path=TEST_FILEPATH)

            deleted_path = delete_from_cache(TEST_MODEL_HANDLE, path=TEST_FILEPATH)

            self.assertEqual(os.path.join(d, EXPECTED_MODEL_SUBPATH), deleted_path)
            self.assertFalse(os.path.exists(os.path.join(get_cached_path(TEST_MODEL_HANDLE), TEST_FILEPATH)))



import json
import os
from http.server import BaseHTTPRequestHandler
from pathlib import Path
from unittest import mock

import requests

import kagglehub
from kagglehub.config import DISABLE_KAGGLE_CACHE_ENV_VAR_NAME
from kagglehub.kaggle_cache_resolver import ATTACH_DATASOURCE_REQUEST_NAME, KAGGLE_CACHE_MOUNT_FOLDER_ENV_VAR_NAME
from tests.fixtures import BaseTestCase

from .utils import create_test_jwt_http_server

INVALID_ARCHIVE_MODEL_HANDLE = "metaresearch/llama-2/pyTorch/bad-archive-variation/1"
VERSIONED_MODEL_HANDLE = "metaresearch/llama-2/pyTorch/13b/1"
LATEST_MODEL_VERSION = 2
UNVERSIONED_MODEL_HANDLE = "metaresearch/llama-2/pyTorch/13b"
TEST_FILEPATH = "config.json"


class KaggleJwtHandler(BaseHTTPRequestHandler):
    def do_HEAD(self) -> None:  # noqa: N802
        self.send_response(200)

    def do_POST(self) -> None:  # noqa: N802
        if self.path.endswith(ATTACH_DATASOURCE_REQUEST_NAME):
            content_length = int(self.headers["Content-Length"])
            request = json.loads(self.rfile.read(content_length))
            model_ref = request["modelRef"]
            version_number = LATEST_MODEL_VERSION
            if "VersionNumber" in model_ref:
                version_number = model_ref["VersionNumber"]

            mount_slug = (
                f"{model_ref['ModelSlug']}/{model_ref['Framework']}/{model_ref['InstanceSlug']}/{version_number}"
            )

            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()

            # Load the files
            cache_mount_folder = os.getenv(KAGGLE_CACHE_MOUNT_FOLDER_ENV_VAR_NAME)
            base_path = f"{cache_mount_folder}/{mount_slug}"
            os.makedirs(base_path, exist_ok=True)

            Path(f"{base_path}/config.json").touch()

            if version_number == LATEST_MODEL_VERSION:
                # The latest version has an extra file.
                Path(f"{base_path}/model.keras").touch()

            # Return the response
            self.wfile.write(
                bytes(
                    json.dumps(
                        {
                            "wasSuccessful": True,
                            "result": {
                                "mountSlug": mount_slug,
                            },
                        }
                    ),
                    "utf-8",
                )
            )
        else:
            self.send_response(404)
            self.wfile.write(bytes(f"Unhandled path: {self.path}", "utf-8"))


# Test cases for the ModelKaggleCacheResolver.
class TestKaggleCacheModelDownload(BaseTestCase):
    def test_unversioned_model_download(self) -> None:
        with create_test_jwt_http_server(KaggleJwtHandler):
            model_path = kagglehub.model_download(UNVERSIONED_MODEL_HANDLE)
            self.assertTrue(model_path.endswith("/2"))
            self.assertEqual(["config.json", "model.keras"], sorted(os.listdir(model_path)))

    def test_versioned_model_download(self) -> None:
        with create_test_jwt_http_server(KaggleJwtHandler):
            model_path = kagglehub.model_download(VERSIONED_MODEL_HANDLE)
            self.assertTrue(model_path.endswith("/1"))
            self.assertEqual(["config.json"], sorted(os.listdir(model_path)))

    def test_versioned_model_download_with_path(self) -> None:
        with create_test_jwt_http_server(KaggleJwtHandler):
            model_file_path = kagglehub.model_download(VERSIONED_MODEL_HANDLE, "config.json")
            self.assertTrue(model_file_path.endswith("config.json"))
            self.assertTrue(os.path.isfile(model_file_path))

    def test_unversioned_model_download_with_path(self) -> None:
        with create_test_jwt_http_server(KaggleJwtHandler):
            model_file_path = kagglehub.model_download(UNVERSIONED_MODEL_HANDLE, "config.json")
            self.assertTrue(model_file_path.endswith("config.json"))
            self.assertTrue(os.path.isfile(model_file_path))

    def test_versioned_model_download_with_missing_file_raises(self) -> None:
        with create_test_jwt_http_server(KaggleJwtHandler):
            with self.assertRaises(ValueError):
                kagglehub.model_download(VERSIONED_MODEL_HANDLE, "missing.txt")

    def test_unversioned_model_download_with_missing_file_raises(self) -> None:
        with create_test_jwt_http_server(KaggleJwtHandler):
            with self.assertRaises(ValueError):
                kagglehub.model_download(UNVERSIONED_MODEL_HANDLE, "missing.txt")

    def test_kaggle_resolver_skipped(self) -> None:
        with mock.patch.dict(os.environ, {DISABLE_KAGGLE_CACHE_ENV_VAR_NAME: "true"}):
            with create_test_jwt_http_server(KaggleJwtHandler):
                # Assert that a ConnectionError is set (uses HTTP server which is not set)
                with self.assertRaises(requests.exceptions.ConnectionError):
                    kagglehub.model_download(VERSIONED_MODEL_HANDLE)

    def test_versioned_model_download_bad_handle_raises(self) -> None:
        with self.assertRaises(ValueError):
            kagglehub.model_download("bad handle")

    def test_versioned_model_download_with_force_download(self) -> None:
        with create_test_jwt_http_server(KaggleJwtHandler):
            model_path = kagglehub.model_download(VERSIONED_MODEL_HANDLE)
            model_path_forced = kagglehub.model_download(VERSIONED_MODEL_HANDLE, force_download=True)

            # Using force_download shouldn't change the expected output of model_download.
            self.assertTrue(model_path_forced.endswith("/1"))
            self.assertEqual(["config.json"], sorted(os.listdir(model_path_forced)))
            self.assertEqual(model_path, model_path_forced)

    def test_versioned_model_download_with_force_download_explicitly_false(self) -> None:
        with create_test_jwt_http_server(KaggleJwtHandler):
            model_path = kagglehub.model_download(VERSIONED_MODEL_HANDLE, force_download=False)
            self.assertTrue(model_path.endswith("/1"))
            self.assertEqual(["config.json"], sorted(os.listdir(model_path)))



__version__ = "0.2.2"

import kagglehub.logging  # configures the library logger.
from kagglehub import colab_cache_resolver, http_resolver, kaggle_cache_resolver, registry
from kagglehub.auth import login
from kagglehub.models import model_download, model_upload

registry.model_resolver.add_implementation(http_resolver.ModelHttpResolver())
registry.model_resolver.add_implementation(kaggle_cache_resolver.ModelKaggleCacheResolver())
# TODO(b/313706281): Register dataset resolvers here.
registry.model_resolver.add_implementation(colab_cache_resolver.ModelColabCacheResolver())



import logging

from kagglehub.config import get_log_verbosity


def _configure_logger() -> None:
    library_name = __name__.split(".")[0]  # i.e. "kagglehub"
    library_logger = logging.getLogger(library_name)
    library_logger.addHandler(logging.StreamHandler())
    # Disable propagation of the library log outputs.
    # This prevents the same message again from being printed again if a root logger is defined.
    library_logger.propagate = False
    library_logger.setLevel(get_log_verbosity())


_configure_logger()



import logging
import os
import time
from typing import Optional

from kagglehub.clients import (
    DEFAULT_CONNECT_TIMEOUT,
    KaggleJwtClient,
)
from kagglehub.config import is_kaggle_cache_disabled
from kagglehub.env import is_in_kaggle_notebook
from kagglehub.exceptions import BackendError
from kagglehub.handle import ModelHandle
from kagglehub.resolver import Resolver

KAGGLE_CACHE_MOUNT_FOLDER_ENV_VAR_NAME = "KAGGLE_CACHE_MOUNT_FOLDER"
ATTACH_DATASOURCE_REQUEST_NAME = "AttachDatasourceUsingJwtRequest"
# b/312965617: Using a longer timeout for this RPC.
ATTACH_DATASOURCE_READ_TIMEOUT = 30  # seconds

DEFAULT_KAGGLE_CACHE_MOUNT_FOLDER = "/kaggle/input"


logger = logging.getLogger(__name__)


class ModelKaggleCacheResolver(Resolver[ModelHandle]):
    def is_supported(self, *_, **__) -> bool:  # noqa: ANN002, ANN003
        if is_kaggle_cache_disabled():
            return False

        if is_in_kaggle_notebook():
            return True

        return False

    def __call__(self, h: ModelHandle, path: Optional[str] = None, *, force_download: Optional[bool] = False) -> str:
        if force_download:
            logger.warning("Ignoring invalid input: force_download flag cannot be used in a Kaggle notebook")

        if path:
            logger.info(f"Attaching '{path}' from model '{h}' to your Kaggle notebook...")
        else:
            logger.info(f"Attaching model '{h}' to your Kaggle notebook...")
        client = KaggleJwtClient()
        model_ref = {
            "OwnerSlug": h.owner,
            "ModelSlug": h.model,
            "Framework": h.framework,
            "InstanceSlug": h.variation,
        }
        if h.is_versioned():
            model_ref["VersionNumber"] = str(h.version)

        result = client.post(
            ATTACH_DATASOURCE_REQUEST_NAME,
            {
                "modelRef": model_ref,
            },
            timeout=(DEFAULT_CONNECT_TIMEOUT, ATTACH_DATASOURCE_READ_TIMEOUT),
        )
        if "mountSlug" not in result:
            msg = "'result.mountSlug' field missing from response"
            raise BackendError(msg)

        base_mount_path = os.getenv(KAGGLE_CACHE_MOUNT_FOLDER_ENV_VAR_NAME, DEFAULT_KAGGLE_CACHE_MOUNT_FOLDER)
        cached_path = f"{base_mount_path}/{result['mountSlug']}"

        if not os.path.exists(cached_path):
            # Only print this if the model is not already mounted.
            logger.info(f"Mounting files to {cached_path}...")

        while not os.path.exists(cached_path):
            time.sleep(5)

        if path:
            cached_filepath = f"{cached_path}/{path}"
            if not os.path.exists(cached_filepath):
                msg = (
                    f"'{path}' is not present in the model files. "
                    f"You can access the other files of the attached model at '{cached_path}'"
                )
                raise ValueError(msg)
            return cached_filepath
        return cached_path



import logging
import os
import tarfile
from typing import Optional

from kagglehub.cache import (
    delete_from_cache,
    get_cached_archive_path,
    get_cached_path,
    load_from_cache,
    mark_as_complete,
)
from kagglehub.clients import KaggleApiV1Client
from kagglehub.handle import ModelHandle
from kagglehub.resolver import Resolver

MODEL_INSTANCE_VERSION_FIELD = "versionNumber"

logger = logging.getLogger(__name__)


class ModelHttpResolver(Resolver[ModelHandle]):
    def is_supported(self, *_, **__) -> bool:  # noqa: ANN002, ANN003
        # Downloading files over HTTP is supported in all environments for all handles / path.
        return True

    def __call__(self, h: ModelHandle, path: Optional[str] = None, *, force_download: Optional[bool] = False) -> str:
        api_client = KaggleApiV1Client()

        if not h.is_versioned():
            h.version = _get_current_version(api_client, h)

        model_path = load_from_cache(h, path)
        if model_path and not force_download:
            return model_path  # Already cached
        elif model_path and force_download:
            delete_from_cache(h, path)

        url_path = _build_download_url_path(h)
        out_path = get_cached_path(h, path)

        # Create the intermediary directories
        if path:
            # Downloading a single file.
            os.makedirs(os.path.dirname(out_path), exist_ok=True)
            api_client.download_file(url_path + "/" + path, out_path, h)
        else:
            # Downloading the full archived bundle.
            archive_path = get_cached_archive_path(h)
            os.makedirs(os.path.dirname(archive_path), exist_ok=True)

            # First, we download the archive.
            api_client.download_file(url_path, archive_path, h)

            # Create the directory to extract the archive to.
            os.makedirs(out_path, exist_ok=True)

            if not tarfile.is_tarfile(archive_path):
                msg = "Unsupported archive type."
                raise ValueError(msg)

            # Extract all files to this directory.
            logger.info("Extracting model files...")
            with tarfile.open(archive_path) as f:
                # Model archives are created by Kaggle via the Databundle Worker.
                f.extractall(out_path)

            # Delete the archive
            os.remove(archive_path)

        mark_as_complete(h, path)
        return out_path


def _get_current_version(api_client: KaggleApiV1Client, h: ModelHandle) -> int:
    json_response = api_client.get(_build_get_instance_url_path(h), h)
    if MODEL_INSTANCE_VERSION_FIELD not in json_response:
        msg = f"Invalid GetModelInstance API response. Expected to include a {MODEL_INSTANCE_VERSION_FIELD} field"
        raise ValueError(msg)

    return json_response[MODEL_INSTANCE_VERSION_FIELD]


def _build_get_instance_url_path(h: ModelHandle) -> str:
    return f"models/{h.owner}/{h.model}/{h.framework}/{h.variation}/get"


def _build_download_url_path(h: ModelHandle) -> str:
    return f"models/{h.owner}/{h.model}/{h.framework}/{h.variation}/{h.version}/download"
